# Simulation for a modified dwa algorithm in dynamic environment

import pygame, os, math, time, random, copy
import pygame.gfxdraw
from pygame.locals import *
import time
import pandas as pd
from matplotlib import pyplot as plt
import numpy as np
from sympy import symbols, Eq, solve

from owlready2 import *
from owlready2.reasoning import *
from itertools import product

# Construct the GUI game
pygame.init()

# Set dimensions of game GUI
screen = pygame.display.set_mode([1200, 700])

# Take image as input
img = pygame.image.load()
# Set image as icon
pygame.display.set_icon(img)
pygame.display.set_caption('Movement of USV')
# Set running value
running = True

# Set random seed
random.seed(0)


# ontology_path = 'path_to_your_ontology.owl'
onto = get_ontology(ontology_path).load()


def normalize_1(values):
    min_value = min(values)
    max_value = max(values)
    range_value = max_value - min_value
    # Handle the case where all values are the same
    if range_value == 0:
        return [-1 if x < 0 else 1 for x in
                values]  # or return [0 for x in values] if you prefer zero when all values are equal
    return [2 * (x - min_value) / range_value - 1 for x in values]


def point_to_line_distance(point, line_start, line_end):
    point = np.array(point)
    line_start = np.array(line_start)
    line_end = np.array(line_end)

    line_vec = line_end - line_start
    point_vec = point - line_start

    line_len = np.linalg.norm(line_vec)
    if line_len == 0:
        return np.linalg.norm(point - line_start) 

    line_unit_vec = line_vec / line_len
    projection_length = np.dot(point_vec, line_unit_vec)

    if not np.isscalar(projection_length):
        projection_length = projection_length.item() 
    if projection_length < 0:
        nearest_point = line_start
    elif projection_length > line_len:
        nearest_point = line_end
    else:
        nearest_point = line_start + projection_length * line_unit_vec

    distance = np.linalg.norm(point - nearest_point)
    return distance





def determine_sector(ar):
    # Assuming each sector spans 60 degrees
    if 0 <= ar < 11.25 * math.pi / 180 or 348.75 * math.pi / 180 <= ar < 360 * math.pi / 180:
        return 'Sector I'
    elif 11.25 * math.pi / 180 <= ar < 67.5 * math.pi / 180:
        return 'Sector II'
    elif 67.5 * math.pi / 180 <= ar < 112.5 * math.pi / 180:
        return 'Sector III'
    elif 112.5 * math.pi / 180 <= ar < 247.5 * math.pi / 180:
        return 'Sector IV'
    elif 247.5 * math.pi / 180 <= ar < 292.5 * math.pi / 180:
        return 'Sector V'
    elif 292.5 * math.pi / 180 <= ar < 348.75 * math.pi / 180:
        return 'Sector VI'


def normalize_angle(angle):
    # Normalizes the angle between 0 and 360 degrees
    return angle % 360


def determine_encounter_type(ar, fe):
    if 'Sector I' == determine_sector(ar):
        if 348.75 * math.pi / 180 <= fe <= 360 * math.pi / 180 or 0 * math.pi / 180 <= fe <= 11.25 * math.pi / 180:
            # print("Head-on", "give-way", "give-way")
            return "Head-on", "give-way", "give-way"
        elif 11.25 * math.pi / 180 < fe < 146.25 * math.pi / 180:
            # print("Crossing", "give-way", "stand-on")
            return "Crossing", "give-way", "stand-on"
        elif 146.25 * math.pi / 180 <= fe <= 213.75 * math.pi / 180:
            # print("Overtaking", "give-way", "stand-on")
            return "Overtaking", "give-way", "stand-on"
        elif 213.75 * math.pi / 180 < fe < 348.75 * math.pi / 180:
            # print("Crossing", "stand-on", "give-way")
            return "Crossing", "stand-on", "give-way"
    elif 'Sector II' == determine_sector(ar):
        if 11.25 * math.pi / 180 < fe < 112.5 * math.pi / 180:
            # print( "Crossing", "give-way", "stand-on")
            return "Crossing", "give-way", "stand-on"
        elif 112.5 * math.pi / 180 <= fe <= 180 * math.pi / 180:
            # print("Overtaking", "give-way", "stand-on")
            return "Overtaking", "give-way", "stand-on"
        elif 180 * math.pi / 180 < fe < 360 * math.pi / 180 or 0 * math.pi / 180 <= fe <= 11.25 * math.pi / 180:
            # print("Safe", "keep going", "keep a steady speed and course")
            return "Safe", "keep going", "keep a steady speed and course"
    elif 'Sector III' == determine_sector(ar):
        if 11.25 * math.pi / 180 < fe <= 135 * math.pi / 180:
            # print("Crossing", "give-way", "stand-on")
            return "Crossing", "give-way", "stand-on"
        elif 135 * math.pi / 180 < fe < 180 * math.pi / 180:
            # print("Crossing", "give-way", "stand-on")
            return "Crossing", "quarter lee give-way", "stand-on"
        elif 180 * math.pi / 180 <= fe < 360 * math.pi / 180 or 0 * math.pi / 180 <= fe <= 11.25 * math.pi / 180:
            return "Safe", "keep going", "keep a steady speed and course"
    elif 'Sector IV' == determine_sector(ar):
        if 112.5 * math.pi / 180 <= fe <= 247.5 * math.pi / 180:
            # print("being overtaken", "keep a steady speed and course", "give-way")
            return "being overtaken", "keep a steady speed and course", "give-way"
        elif 247.5 * math.pi / 180 < fe < 360 * math.pi / 180 or 0 * math.pi / 180 < fe < 112.5 * math.pi / 180:
            # print("Safe", "keep going", "keep a steady speed and course")
            return "Safe", "keep going", "keep a steady speed and course"
    elif 'Sector V' == determine_sector(ar):
        if 180 * math.pi / 180 < fe < 225 * math.pi / 180:
            # print("Crossing", "quarter lee stand-on", "give-way")
            return "Crossing", "quarter lee stand-on", "give-way"
        elif 225 * math.pi / 180 <= fe < 360 * math.pi / 180:
            # print("Crossing", "stand-on", "give-way")
            return "Crossing", "stand-on", "give-way"
        elif 0 * math.pi / 180 <= fe <= 180 * math.pi / 180:
            # print("Safe", "keep going", "keep a steady speed and course")
            return "Safe", "keep going", "keep a steady speed and course"
    elif 'Sector VI' == determine_sector(ar):
        if 247.5 * math.pi / 180 < fe < 348.75 * math.pi / 180:
            # print("Crossing", "stand-on", "give-way")
            return "Crossing", "stand-on", "give-way"
        elif 180 * math.pi / 180 <= fe <= 247.5 * math.pi / 180:
            # print("Overtaking", "give-way", "stand-on")
            return "Overtaking", "give-way", "stand-on"
        elif 348.75 * math.pi / 180 <= fe < 360 * math.pi / 180 or 0 * math.pi / 180 <= fe < 180 * math.pi / 180:
            # print("Safe", "keep going", "keep a steady speed and course")
            return "Safe", "keep going", "keep a steady speed and course"
    else:
        return "Unknown", "Unknown", "Unknown"


###################################################################

class Ship:
    def __init__(self, screen, k, u0, v0, color=(0, 76, 153), ship_length=1.25, ship_width=0.3):
        # Initialize pygame
        pygame.init()

        # Set screen dimensions from the screen object
        self.WIDTH, self.HEIGHT = screen.get_size()
        # Set scaling and offset parameters
        self.k = k
        self.u0 = u0
        self.v0 = v0
        self.GRID_SIZE = 50  # pixels between each grid line
        self.frame_counter = 0

        # Take image as input
        img = pygame.image.load(
            'C:/Users/rongxinsong/OneDrive - Delft University of Technology/TU Delft PhD/TU Delft/PhD project route/500ppi/500ppi/USV2.png')

        # Set image as icon
        pygame.display.set_icon(img)
        pygame.display.set_caption("Path planner with Knowledge Map-driven DWA Algorithm")

        self.screen = screen
        self.expansion_radius = 2 * ship_length
        self.ship_color = color
        self.ship_length = ship_length
        self.ship_width = ship_width
        self.ship_shape = self.define_ship_shape()
        self.collision_avoidance_status = {}

    def update_collision_avoidance_status(self, target_ship_id, status):
        self.collision_avoidance_status[target_ship_id] = status

    def define_ship_shape(self):
        # Parameters
        trapezoid_height = self.ship_width / 10
        cab_length = self.ship_length / 10
        cab_width = self.ship_width / 1.8

        # Ellipse parameters
        a = self.ship_length / 2.5
        b = self.ship_width / 2

        # Ship shape vertices (centered around origin)
        ship_vertices = np.array([
            [-self.ship_length / 2, -self.ship_width / 2],
            [self.ship_length / 2, -self.ship_width / 2],
            [self.ship_length / 2, self.ship_width / 2],
            [-self.ship_length / 2, self.ship_width / 2]
        ])

        trapezoid_vertices = np.array([
            [-self.ship_length / 2 - trapezoid_height, -self.ship_width / 2 + self.ship_width / 25],
            [-self.ship_length / 2, -self.ship_width / 2],
            [-self.ship_length / 2, self.ship_width / 2],
            [-self.ship_length / 2 - trapezoid_height, self.ship_width / 2 - self.ship_width / 25]
        ])

        cab_vertices = np.array([
            [-cab_length / 2, -cab_width / 2],
            [cab_length / 2, -cab_width / 2],
            [cab_length / 2, cab_width / 2],
            [-cab_length / 2, cab_width / 2]
        ])

        return ship_vertices, trapezoid_vertices, cab_vertices, (a, b)

    def draw_ship(self, x0, y0, theta,
                  draw_ellipse_half=True,
                  draw_trapezoid_sides=[True, False, True, True],
                  draw_rect_sides=[True, False, True, False]):

        theta_rotation = math_to_nautical_angle(theta)

        R = np.array([
            [np.sin(theta_rotation), np.cos(theta_rotation)],
            [np.cos(theta_rotation), -np.sin(theta_rotation)]
        ])

        def transform_and_draw(vertices, color, sides=None, close_path=True):
            if sides is None:
                sides = [True] * len(vertices)
            # Scale, rotate, and translate
            transformed_vertices = vertices @ R.T + np.array([x0, y0])
            # transformed_vertices = vertices * self.scale_factor @ R.T + np.array([x0 * self.scale_factor, y0 * self.scale_factor])

            # Convert to pygame coordinates
            pygame_vertices = [(self.u0 + x * self.k, self.v0 - y * self.k) for x, y in transformed_vertices]
            # pygame_vertices = [(x, self.HEIGHT - y) for x, y in transformed_vertices]

            assert len(sides) == len(vertices), "Sides and vertices lists must have the same length!"
            # Manually draw each segment for the ellipse
            if not close_path:
                for i in range(len(pygame_vertices) - 1):
                    pygame.draw.line(self.screen, color, pygame_vertices[i], pygame_vertices[i + 1], 2)
            else:
                # If specific sides are defined, draw them selectively
                if sides:
                    for i in range(len(vertices)):
                        if sides[i]:
                            pygame.draw.line(self.screen, color, pygame_vertices[i],
                                             pygame_vertices[(i + 1) % len(vertices)], 2)
                else:
                    pygame.draw.lines(self.screen, color, True, pygame_vertices, 2)

        # Unpack ship parts
        ship_vertices, trapezoid_vertices, cab_vertices, ellipse_dims = self.ship_shape
        a, b = ellipse_dims

        if draw_ellipse_half:
            # Define the top half of the ellipse in the ship's local coordinate system
            num_points = 100
            theta_vals = np.linspace(-math.pi / 2, math.pi / 2, num_points)[:-1]  # Exclude the last point
            ellipse_points_local = np.array(
                [[a + a * np.cos(theta_val), b * np.sin(theta_val)] for theta_val in theta_vals])

            # Use the existing function to transform and draw this ellipse segment
            # print(f'real-time theta:{np.deg2rad(theta)}')
            transform_and_draw(ellipse_points_local, self.ship_color, close_path=False)

        # Draw specific sides
        transform_and_draw(ship_vertices, self.ship_color, draw_rect_sides)
        transform_and_draw(trapezoid_vertices, self.ship_color, draw_trapezoid_sides)

    def draw_arrow(self, x0, y0, theta, color=(255, 215, 0)):
        """Draws an arrow indicating the direction of the ship."""
        # Convert theta to radians and adjust for Pygame coordinates
        # theta_radians = np.deg2rad(-theta + 90)
        # theta_radians = np.deg2rad(theta)
        # theta_radians = np.deg2rad(math_to_nautical_angle(np.deg2rad(theta)))
        theta_radians = math_to_nautical_angle(theta)

        length = 1.5
        # Calculate the end point of the arrow
        end_x = x0 + length * np.sin(theta_radians)
        end_y = y0 + length * np.cos(theta_radians)

        # Convert starting and ending points to screen coordinates
        screen_start_x = self.u0 + self.k * x0
        screen_start_y = self.v0 - self.k * y0
        screen_end_x = self.u0 + self.k * end_x
        screen_end_y = self.v0 - self.k * end_y

        # Draw the main line of the arrow
        pygame.draw.line(self.screen, color, (screen_start_x, screen_start_y), (screen_end_x, screen_end_y), 2)

        # Calculate points for the arrowhead
        arrow_angle = math.pi / 1.2  # 30 degrees for arrowhead
        arrowhead_length = 10
        left_end_x = screen_end_x + arrowhead_length * np.sin(theta_radians - arrow_angle)
        left_end_y = screen_end_y - arrowhead_length * np.cos(theta_radians - arrow_angle)
        right_end_x = screen_end_x + arrowhead_length * np.sin(theta_radians + arrow_angle)
        right_end_y = screen_end_y - arrowhead_length * np.cos(theta_radians + arrow_angle)

        # Draw the arrowhead
        pygame.draw.line(self.screen, color, (screen_end_x, screen_end_y), (left_end_x, left_end_y), 2)
        pygame.draw.line(self.screen, color, (screen_end_x, screen_end_y), (right_end_x, right_end_y), 2)


    def draw_axes_and_grid_enhanced(self, k, grid_color=(220, 220, 220), axis_color=(50, 50, 50)):
        # Initialize font for labels
        font = pygame.font.Font(None, 24)
        FONT_COLOR = (80, 80, 80)  # Moderate grey color for font

        # Fixed origin at (0,0) in screen coordinates
        u0 = int(1.2 * k)
        v0 = self.HEIGHT - int(1.2 * k)
        grid_size = k  # 1-to-1 scale for grid size

        # Draw the x and y axis with arrows
        pygame.draw.line(self.screen, axis_color, (u0, 0), (u0, self.HEIGHT), 2)
        pygame.draw.line(self.screen, axis_color, (0, v0), (self.WIDTH, v0), 2)

        pygame.draw.polygon(self.screen, axis_color, [(self.WIDTH - 10, v0 - 5), (self.WIDTH, v0),
                                                      (self.WIDTH - 10, v0 + 5)])  # Arrow for x-axis
        pygame.draw.polygon(self.screen, axis_color, [(u0 - 5, 10), (u0, 0), (u0 + 5, 10)])  # Arrow for y-axis

        # Calculate the position where the "X" and "Y" labels should be drawn
        label_x_pos = self.WIDTH - 20
        label_y_pos = 20

        # Draw "X" and "Y" labels
        label_x = font.render("E", True, FONT_COLOR)
        label_y = font.render("N", True, FONT_COLOR)
        self.screen.blit(label_x, (label_x_pos, v0 - 20))  # Position for "X" label
        self.screen.blit(label_y, (u0 + 10, label_y_pos))  # Position for "Y" label

        # Draw grid with labels based on fixed intervals starting from 0
        # Draw X axis labels
        for x in range(u0, self.WIDTH, grid_size):
            pygame.draw.line(self.screen, grid_color, (x, 0), (x, self.HEIGHT), 1)
            label = font.render(f"{(x - u0) // k}", True, FONT_COLOR)
            self.screen.blit(label, (x + 5, v0 + 10))

        # Draw Y axis labels
        for y in range(v0, 0, -grid_size):
            pygame.draw.line(self.screen, grid_color, (0, y), (self.WIDTH, y), 1)
            label = font.render(f"{(v0 - y) // k}", True, FONT_COLOR)
            self.screen.blit(label, (u0 - 20, y - 10))

        # Draw the origin label
        label_origin = font.render("0", True, FONT_COLOR)
        self.screen.blit(label_origin, (u0 + 5, v0 + 10))

 
def update_expansion_area(ship_length, u):
    d5 = ship_length
    return d5


class Environment:
    """ Simulation Environment contains moving obstacles
        and agent which performs collision aovidance algorithm.
    """

    def __init__(self, algorithm, num_obstacles, csv_file_path):
        # Set parameters for drawing
        self.PLAYFIELDCORNERS = (0, 0, 30, 30)  # The region we will fill with obstacles
        self.WIDTH, self.HEIGHT = 1200, 700
        self.BORDER_FRACTION = 0.1
        self.effective_width = self.WIDTH * (1 - self.BORDER_FRACTION)
        self.effective_height = self.HEIGHT * (1 - self.BORDER_FRACTION)
        self.sim_width = 15  # Replace with actual width
        self.sim_height = 10  # Replace with actual height
        self.paused = False
        self.k = min(self.effective_width / self.sim_width, self.effective_height / self.sim_height) - 40

        self.u0 = 0
        self.v0 = self.HEIGHT

        self.screen = pygame.display.set_mode((self.WIDTH, self.HEIGHT))

        self.size = [self.WIDTH, self.HEIGHT]
        self.black = (20, 20, 40)
        self.lightblue = (0, 120, 255)
        self.darkblue = (0, 40, 160)
        self.red = (255, 100, 0)
        self.white = (255, 255, 255)
        self.blue = (0, 0, 255)
        self.grey = (70, 70, 70)
        self.ocean = (102, 153, 204)
        self.ocean1 = (32, 178, 170)
        self.ocean2 = (240, 255, 255)
        self.ocean3 = (25, 25, 112)
        self.ocean4 = (64, 224, 208)
        self.ocean5 = (0, 105, 148)
        self.ship = Ship(self.screen, self.k, self.u0, self.v0)
        # self.dwa = NewDWA(self)

        pygame.init()
        # Initialise Pygame display screen
        self.screen = pygame.display.set_mode(self.size)
        # This makes the normal mouse pointer invisible/visible (0/1) in graphics window
        pygame.mouse.set_visible(1)

        # Our algorithm
        self.alg = algorithm
        # Simulatation time interval
        self.dt = 0.2
        # Initiate obstacles
        self.OBSTACLE_RADIUS = 0.20
        self.OBSTACLE_MAX_VEL = 0.2
        self.init_obstacle_num = num_obstacles
        # Initiate USV's goal
        self.goal = (self.PLAYFIELDCORNERS[2] - 1, self.PLAYFIELDCORNERS[3] - 1)
        self.load_obstacle_positions_from_csv(csv_file_path)

        # Initiate pose of USV
        self.init_x = self.PLAYFIELDCORNERS[0] + 3
        self.init_y = self.PLAYFIELDCORNERS[1] + 3
        self.init_theta = math.pi / 2

        self.init_pose = (self.init_x, self.init_y, self.init_theta)
        self.config = (self.dt, self.k, self.u0, self.v0, self.OBSTACLE_RADIUS)
        # Global parameters for evaluate algorithms
        self.sim_times = 0  # num of simulation times
        self.collision_times = 0  # num of collision times during all simulations
        self.avg_tg = 0
        self.avg_distance_travelled = 0
        self.tg_vec = []
        self.vl_vec = []
        self.vs_vec = []
        self.vr_vec = []
        self.distance_travelled_vec = []
        # Set moving things
        self.reset()

    def reset(self):
        """ Reset parameters for next simulation """
        # Reset obstacles
        self.obstacles = []  # obstalces list

        ship_colors = [(0, 0, 0), (0, 0, 0), (0, 0, 0), (0, 0, 0)]  # 定义不同的颜色
        ship_sizes = [(1.25, 0.3), (1.25, 0.3), (1.25, 0.3), (1.25, 0.3)]  # 定义不同的尺寸 (长度, 宽度)

        shipType = ['engaged_in_fishing', 'powerdriven_vessel', 'restricted_in_her_ability_to_maneuver',
                    'sailing_vessel', 'seaplane', 'vessel_constrained_by_her_draught', 'vessel_ont_under_command',
                    'WingInGround_craft']
        fixed_parameters = [
            ('TS1', 5, 18, 0.2, -0.45, shipType[2]),
            # Vessel 1 - starting at (5, 10), moving East at 1 unit per time, heading East
            # ('TS2', 3, 22, 0.3, -0.2, shipType[1]),  # Vessel 2 - starting at (10, 5), moving North at 1 unit per time, heading North
            # # (3, 3, 28, 0.5, -0.1, 100/180 * math.pi),  # Vessel 3 - starting at (15, 10), moving West at 1 unit per time, heading West
            # ('TS4', 28, 18, -0.15, 0.1, shipType[0]),  # Vessel 4 - starting at (10, 15), moving South at 1 unit per time, heading South
            # ('TS5', 30, 30, -0.1, -0.15, shipType[-2])
            # Vessel 5 - starting at (20, 20), moving Northeast at sqrt(2) units per time, heading Northeast
        ]


        # Use the fixed parameters to set the properties of each obstacle
        for i, params in enumerate(fixed_parameters):
            id, pos_x, pos_y, vx, vy, shipT = params
            heading = math_to_nautical_angle(math.atan2(vy, vx))
            # print(math.degrees(heading))

            # Select color and size for the ship (unchanged)
            color = ship_colors[i % len(ship_colors)]
            size = ship_sizes[i % len(ship_sizes)]
            ship_length, ship_width = size

            # Create a dictionary for each obstacle with its properties
            obstacle = {
                'id': id,
                'pos_x': pos_x,
                'pos_y': pos_y,
                'vx': vx,
                'vy': vy,
                'color': color,
                'size': (ship_length, ship_width),
                'heading': heading,
                'shiptype': shipT
            }

            self.obstacles.append(obstacle)

        # Reset algorithm
        self.agent = self.alg(self.init_pose, self.config)
        self.agent.ship_length = 1.25
        # Reset sim flag
        self.sim_over = False  # indicates whether a simulation is over or not
        # Reset collision flag
        self.collision_happened = False
        # Reset time to reach the goal
        self.time_to_goal = 0.0
        # Reset distance_travelled
        self.distance_travelled = 0
        # Record USV's history positions and paths
        self.history_positions = []
        # self.theta = []
        self.history_vl_vec = []
        self.history_vs_vec = []
        self.history_vr_vec = []

    def load_obstacle_positions_from_csv(self, file_path):
        df = pd.read_csv(file_path)
        self.obstacle_positions = list(zip(df['X_Tar'].values, df['Y_Tar'].values))

    def move_obstacles(self, deltatime):
        # deltatime = self.dt
        """ Update locations and velocties of moving obstacles"""
        for idx in range(len(self.obstacles)):
            # Update x coordinate
            self.obstacles[idx]['pos_x'] += self.obstacles[idx]['vx'] * deltatime
            if self.obstacles[idx]['pos_x'] < self.PLAYFIELDCORNERS[0]:
                self.obstacles[idx]['vx'] = -self.obstacles[idx]['vx']
            if self.obstacles[idx]['pos_x'] > self.PLAYFIELDCORNERS[2]:
                self.obstacles[idx]['vx'] = -self.obstacles[idx]['vx']
            # Update y coordinate
            self.obstacles[idx]['pos_y'] += self.obstacles[idx]['vy'] * deltatime
            if self.obstacles[idx]['pos_y'] < self.PLAYFIELDCORNERS[1]:
                self.obstacles[idx]['vy'] = -self.obstacles[idx]['vy']
            if self.obstacles[idx]['pos_y'] > self.PLAYFIELDCORNERS[3]:
                self.obstacles[idx]['vy'] = -self.obstacles[idx]['vy']

            # Calculate heading from velocities
            if self.obstacles[idx]['vx'] == 0 and self.obstacles[idx]['vy'] == 0:
                # If there is no movement, the heading is not defined
                # Assign a default value or continue without heading
                self.obstacles[idx]['heading'] = 0  # Or some default value
            else:
                # Calculate heading (in degrees from North direction)
                self.obstacles[idx]['heading'] = math_to_nautical_angle(
                    math.atan2(self.obstacles[idx]['vy'], self.obstacles[idx]['vx']))

    def draw_obstacles(self):
        """ Draw circular obstalces on screen """
        for (idx, obstacle) in enumerate(self.obstacles):
            color = self.lightblue
            pygame.gfxdraw.aacircle(self.screen,
                                    int(self.u0 + self.k * obstacle[0]), int(self.v0 - self.k * obstacle[1]),
                                    int(self.k * self.OBSTACLE_RADIUS + 1), color)

            steps = 10
            linestart = (self.u0 + self.k * obstacle[0], self.v0 - self.k * obstacle[1])
            maxPathLen = self.OBSTACLE_MAX_VEL * steps * self.dt
            dx = obstacle[2] * steps * self.dt
            dy = obstacle[3] * steps * self.dt
            dx, dy = dx / math.sqrt(dx ** 2 + dy ** 2) * maxPathLen, dy / math.sqrt(dx ** 2 + dy ** 2) * maxPathLen
            lineend = (self.u0 + self.k * (obstacle[0] + dx), self.v0 - self.k * (obstacle[1] + dy))
            pygame.draw.line(self.screen, (200, 0, 0), linestart, lineend, 5)

    def draw_obstacles(self):
        """ Draw ships with varying colors and sizes on screen """
        for obstacle in self.obstacles:
            pos_x, pos_y = obstacle['pos_x'], obstacle['pos_y']
            vx, vy = obstacle['vx'], obstacle['vy']
            color = obstacle['color']
            theta = obstacle['heading']
            ship_length, ship_width = obstacle['size']

            # Calculate the dynamic expansion radius
            expansion_radius = update_expansion_area(ship_length, vx)

            ship = Ship(self.screen, self.k, self.u0, self.v0, color, ship_length, ship_width)
            draw_theta = nautical_to_math_angle(theta)
            ship.draw_ship(pos_x, pos_y, draw_theta)

            # Draw the ship's direction arrow
            # ship.draw_arrow(pos_x, pos_y, draw_theta)
            # Draw the dynamic expansion area around the ship
            pygame.draw.circle(self.screen, (255, 100, 100),
                               (int(self.u0 + self.k * pos_x), int(self.v0 - self.k * pos_y)),
                               int((self.k) * expansion_radius), 1)

    def draw_goal(self):
        """ Draw the goal of USV on screen """
        color = self.red
        goal = self.goal
        pygame.draw.circle(self.screen, color, (int(self.u0 + self.k * goal[0]), int(self.v0 - self.k * goal[1])),
                           int(self.k * self.OBSTACLE_RADIUS), 0)

    def normalize_degrees(self, degrees):
        degrees %= 360 
        return degrees

    def draw_USV(self):
        # degree_theta = math_to_nautical_angle(self.agent.theta)
        # degree_theta1 = self.normalize_degrees(degree_theta)
        # print(f'degree theta:{np.deg2rad(degree_theta1)}')
        self.ship.draw_ship(self.agent.x, self.agent.y, self.agent.theta, self.ship.ship_shape)
        self.ship.draw_expansion_area(self.agent.x, self.agent.y, self.agent.ship_length)
        self.ship.draw_expansion_area(self.agent.x, self.agent.y, 3 * self.agent.ship_length, color=(34, 139, 34))
        self.ship.draw_expansion_area(self.agent.x, self.agent.y, 4 * self.agent.ship_length, color=(112, 128, 144))
        self.ship.draw_arrow(self.agent.x, self.agent.y, self.agent.theta, color=(112, 128, 144))

    def draw_history_trajectory(self):
        """ Draw trajectory of moving USV """
        color = self.grey
        # Draw locations
        for pos in self.history_positions:
            pygame.draw.circle(self.screen, color, (int(self.u0 + self.k * pos[0]), int(self.v0 - self.k * pos[1])), 3,
                               0)

    def draw_predicted_tracjetory(self, predicted_path_to_draw, optimal_path=None):
        # Draw paths(straight lines or arcs)
        for path in predicted_path_to_draw:
            color = (0, 200, 0)  # Default color for non-optimal paths
            # color =(255, 0, 0)
            if path == optimal_path:
                color = (255, 0, 0)  # Optimal path color
            if path[0] == 0:  # Straight line
                straightpath = path[1]
                linestart = (self.u0 + self.k * self.agent.x, self.v0 - self.k * self.agent.y)
                lineend = (
                self.u0 + self.k * (self.agent.x + straightpath * math.sin(math_to_nautical_angle(self.agent.theta))),
                self.v0 - self.k * (self.agent.y + straightpath * math.cos(math_to_nautical_angle(self.agent.theta))))
                pygame.draw.line(self.screen, color, linestart, lineend, 3)
                # pygame.gfxdraw.line(self.screen, linestart[0], linestart[1], lineend[0], lineend[1], (0, 200, 0))
            if path[0] == 1:  # Rotation, nothing to draw
                pass
            if path[0] == 2:  # General case: circular arc
                # path[2] and path[3] are start and stop angles for arc but they need to be in the right order to pass
                if (path[3] > path[2]):
                    startangle = path[2]
                    stopangle = path[3]
                else:
                    startangle = path[3]
                    stopangle = path[2]
                # Pygame arc doesn't draw properly unless angles are positive
                if (startangle < 0):
                    startangle += 2 * math.pi
                    stopangle += 2 * math.pi
                if (path[1][1][0] > 0 and path[1][0][0] > 0 and path[1][1][1] > 1):
                    # print (path[1], startangle, stopangle)
                    pygame.draw.arc(self.screen, color, path[1], startangle, stopangle, 3)

    def check_collision(self):
        """ Check if collision happened between USV and moving obstacles """
        for obstacle in self.obstacles:
            ox, oy = obstacle['pos_x'], obstacle['pos_y']
            vx, vy = obstacle['vx'], obstacle['vy']
            ship_length = obstacle['size'][0]
            # Calculate dynamic expansion radius for the obstacle
            own_expansion_radius = update_expansion_area(self.agent.ship_length, self.agent.linear_vel)

            tar_expansion_radius = update_expansion_area(ship_length, vx)

            dist = math.sqrt((ox - self.agent.x) ** 2 + (oy - self.agent.y) ** 2)

            if dist < own_expansion_radius + tar_expansion_radius:  # Collision happened
                self.collision_times += 1
                self.collision_happened = True

    def run(self):
        """ Do simulation """
        if self.sim_over == True:
            self.reset()
        while self.sim_over == False:

            # Start simulation
            self.time_to_goal += self.dt
            self.distance_travelled += np.sqrt((self.agent.linear_vel) ** 2 + (self.agent.lateral_vel) ** 2) * self.dt
            predicted_path_to_draw = []
            # Save USV's locations for display of trail
            self.history_positions.append((self.agent.x, self.agent.y))
            # self.theta.append(self.agent.theta)
            # Planning velocities and path
            predicted_path_to_draw, optimal_path, best_action_index, cost_dict, vLchosen, vSchosen, vRchosen, text, warning_condition = self.agent.planning(
                self.goal, self.obstacles)
            # print(cost_dict)
            best_path_costs = [
                (cost_dict['Dist goal'])[best_action_index],
                (cost_dict['Dist obstacle'])[best_action_index],
                (cost_dict['Path keeping'])[best_action_index],
                (cost_dict['Time to_goal'])[best_action_index],
                (cost_dict['Heading stability'])[best_action_index],
                (cost_dict['COLREGs adherence'])[best_action_index],
            ]

  

            self.history_vl_vec.append(vLchosen)
            self.history_vs_vec.append(vSchosen)
            self.history_vr_vec.append(vRchosen)
            # print(f'optimal_path:{optimal_path}')
            # Visualization
            self.draw_frame(predicted_path_to_draw, optimal_path, best_action_index, best_path_costs, weights,
                            cost_names, text, warning_condition)
            # Check collison
            self.check_collision()
            dist_to_goal = math.sqrt((self.agent.x - self.goal[0]) ** 2 + (self.agent.y - self.goal[1]) ** 2)
            if self.collision_happened == True:
                self.sim_over = True
                self.sim_times += 1
                print('#{} \t Failure \t [tg:  None] \t  [total collision times:{}]'.format(self.sim_times,
                                                                                            self.collision_times))
                # time.sleep(1)
                break
            elif round(dist_to_goal, 3) < self.agent.USV_RADIUS:
                self.sim_over = True
                self.sim_times += 1
                self.tg_vec.append(self.time_to_goal)
                self.vl_vec.extend(self.history_vl_vec)
                self.vs_vec.extend(self.history_vs_vec)
                self.vr_vec.extend(self.history_vr_vec)
                # self.theta.extend(self.theta)
                self.distance_travelled_vec.append(self.distance_travelled)
                print('#{} \t Success \t [tg:  {:.2f}] \t [total collision times:{}]'.format(self.sim_times,
                                                                                             self.time_to_goal,
                                                                                             self.collision_times))
                break
            else:
                # Continue simualtion
                self.move_obstacles(self.dt)  # Move obstacles
                self.agent.move_USV()  # Move USV


class NewDWA:
    """ Collision avoidance algorithm """

    def __init__(self, init_pose=(0, 0, 0), config=(0.10, 0, 0, 0)):
        # parameters of USV
        self.USV_RADIUS = 0.10
        # Linear velocity limits
        self.MAX_VEL_LINEAR = 0.58  # ms^(-1) max speed of each wheel
        self.MAX_ACC_LINEAR = 0.09  # ms^(-2) max rate we can change speed of each wheel

        self.MAX_VEL_LATERAL = 0.3  # maximum lateral speed
        self.MAX_ACC_LATERAL = 0.14  # maximum rate we can change lateral speed

        # Angular velocity limits
        self.MAX_VEL_ANGULAR = 0.85
        self.MAX_ACC_ANGULAR = 0.55

        # Current linear velocity and angular velocity
        self.linear_vel = 0.0
        self.lateral_vel = 0.0
        self.angular_vel = 0.0
        self.init_pose = init_pose
        # Current positions
        self.x, self.y, self.theta = init_pose

        # Parameters for prediction trajectory
        self.dt, self.k, self.u0, self.v0, self.OBSTACLE_RADIUS = config

        self.STEPS_AHEAD_TO_PLAN = STEPS_AHEAD_TO_PLAN
        self.TAU = self.dt * self.STEPS_AHEAD_TO_PLAN

        # Safe distance between USV and closest obstacle after minus USV's radius and obstacle's radius
        # self.SAFE_DIST = self.MAX_VEL_LINEAR ** 2 / (2 * self.MAX_ACC_LINEAR)
        self.SAFE_DIST = (2 * 1.25)
        # self.SAFE_dcpa = (2 * 1.25)
        self.collision_avoidance_status = {}

        # Weights for predicted trajectory evaluation
        self.FORWARDWEIGHT = FORWARDWEIGHT
        self.OBSTACLEWEIGHT = OBSTACLEWEIGHT
        self.PATHKEEPINGWEIGHT = PATHKEEPINGWEIGHT
        self.TIMETOGOALWEIGHT = TIMETOGOALWEIGHT
        self.HEADINGSTABILITYWEIGHT = HEADINGSTABILITYWEIGHT
        self.COLREGSWEIGHT = COLREGSWEIGHT
        # self.TCPAWEIGHT = 12

    def calculate_heading_difference(self, start, goal, predict_theta):
        predict_theta_nautical = math_to_nautical_angle(predict_theta)
        # Calculate the desired heading based on the start and goal positions
        desired_heading = math_to_nautical_angle(math.atan2(goal[1] - start[1], goal[0] - start[0]))
        # print(f'desired_heading:{np.rad2deg(desired_heading)}')
        # desired_heading_nautical = math_to_nautical_angle(desired_heading)

        # print(f'desired_heading:{desired_heading}')

        # Calculate the difference between the desired heading and the ship's current heading
        heading_difference = desired_heading - predict_theta_nautical

        # Normalize the difference to the range [-pi, pi]
        while heading_difference > math.pi:
            heading_difference -= 2 * math.pi
        while heading_difference < -math.pi:
            heading_difference += 2 * math.pi
        return heading_difference


        return new_x, new_y, nautical_to_math_angle(new_theta), path



    def calculateObstacleWeight(self, obstacle):
        # Example: weight based on obstacle size and speed
        size_weight = obstacle['size']  # Example attribute
        speed_weight = math.sqrt(obstacle['vx'] ** 2 + obstacle['vy'] ** 2)
        return size_weight + speed_weight

    def calculate_angle_to_closest_obstacle(self, predict_x, predict_y, USV_orientation, obstacles):

        closest_angle_diff = float('inf')
        for obstacle in obstacles:
            dx = obstacle['pos_x'] - predict_x
            dy = obstacle['pos_y'] - predict_y

            # Calculate angle from USV to obstacle
            angle_to_obstacle = math_to_nautical_angle(math.atan2(dy, dx))

            # Calculate angle difference
            angle_diff = (angle_to_obstacle - USV_orientation)

            # Find smallest angle difference
            if abs(angle_diff) < abs(closest_angle_diff):
                closest_angle_diff = angle_diff

        return closest_angle_diff

    def get_velocity_range_intersection(self, vl_bounds, vs_bounds, vr_bounds):
        vl_min, vl_max = max(-self.MAX_VEL_LINEAR, vl_bounds[0]), min(self.MAX_VEL_LINEAR, vl_bounds[1])
        vs_min, vs_max = max(-self.MAX_VEL_LATERAL, vs_bounds[0]), min(self.MAX_VEL_LATERAL, vs_bounds[1])
        vr_min, vr_max = max(-self.MAX_VEL_ANGULAR, vr_bounds[0]), min(self.MAX_VEL_ANGULAR, vr_bounds[1])
        return (vl_min, vl_max), (vs_min, vs_max), (vr_min, vr_max)

    def enhanced_colregs_compliance(self, distance, encounter_type, action_own, pos_x, pos_y, predict_x, predict_y,
                                    vLpossible, vSpossible, vRpossible, m1=3, m2=1.5):
        warning_condition = True
        text = ' '
        colregs_cost1 = 0
        if encounter_type == "Crossing":
            if action_own == "give-way":
                # if distance < 4 * 1.25:
                warning_condition = True
                text = '-------------WARNING!!---turn to starboard side to give way!---------------'

                colregs_cost1 += max(0, -vRpossible)  # Lower cost for right turn, higher cost for left turn or straight
            elif action_own == "stand-on":
                warning_condition = True
                # text = '-------------WARNING!!keep course and speed!---------------'
                if distance > m1 * 1.25:
                    warning_condition = True
                    text = '---------I WARNING!!---TS give way and OS keep course and speed-----------'
                    colregs_cost1 -= abs(vLpossible - self.linear_vel) + abs(vSpossible - self.lateral_vel) + abs(
                        vRpossible)
                elif distance <= m1 * 1.25:

                    text = '---------II WARNING!!!--TS out! ownship should avoid collisions-----------'
                    colregs_cost1 += max(0, -vRpossible) 

        elif encounter_type == "Head-on":
            # For head-on situations
            if action_own == "give-way":
                # Encourage turning to starboard
                colregs_cost1 += max(0, -vRpossible)

        return colregs_cost1, text, warning_condition

    def calculate_DCPA_TCPA(self, predict_x, predict_y, vLpossible, vSpossible, predict_theta, obstacle):

        predict_theta_nautical = math_to_nautical_angle(predict_theta)
        relative_velocity_x = (
                vLpossible * math.sin(predict_theta_nautical) + vSpossible * math.cos(predict_theta_nautical) -
                obstacle['vx'])
        relative_velocity_y = (
                vLpossible * math.cos(predict_theta_nautical) - vSpossible * math.sin(predict_theta_nautical) -
                obstacle['vy'])
        relative_speed = math.sqrt(relative_velocity_x ** 2 + relative_velocity_y ** 2)

        # Initial separation
        dx = obstacle['pos_x'] - predict_x
        dy = obstacle['pos_y'] - predict_y
        initial_separation = math.sqrt(dx ** 2 + dy ** 2)

        # Angle between relative velocity and line of sight
        angle = math.acos(
            (dx * relative_velocity_x + dy * relative_velocity_y) / (initial_separation * relative_speed + 1e-10))

        # Calculate TCPA
        TCPA = initial_separation * math.cos(angle) / (relative_speed + 1e-10)

        # Calculate DCPA
        DCPA = math.sqrt(initial_separation ** 2 - (TCPA * relative_speed) ** 2)

        return DCPA, TCPA

    # Helper method to check if collision is imminent
    def is_collision_imminent(self, own_ship_state, obstacle, safe_DCPA, safe_TCPA):
        DCPA, TCPA = self.calculate_DCPA_TCPA(own_ship_state['pos_x'], own_ship_state['pos_y'],
                                              own_ship_state['vL'], own_ship_state['vS'],
                                              own_ship_state['theta'], obstacle)
        return DCPA < safe_DCPA or abs(TCPA) < safe_TCPA

    def refine_dcpa_tcpa_handling(self, DCPA, TCPA, DCPA_THRESHOLD=3, TCPA_THRESHOLD=10):
        """ Refine the handling of DCPA and TCPA values. """
        if DCPA < DCPA_THRESHOLD and 0 < TCPA < TCPA_THRESHOLD:
            return True  # Collision is imminent
        return False  # No immediate danger


    def planning(self, goal, obstacles):

        bestBenefit = -100000
        # Range of possible motions: each of vL and vR could go up or down a bit
        sample_num = 8

        aL_max = self.MAX_ACC_LINEAR
        aL_min = -self.MAX_ACC_LINEAR
        aS_max = self.MAX_ACC_LATERAL
        aS_min = -self.MAX_ACC_LATERAL
        aR_max = self.MAX_ACC_ANGULAR
        aR_min = -self.MAX_ACC_ANGULAR

        aL_max_adjusted = min((self.MAX_VEL_LINEAR - self.linear_vel) / self.dt, aL_max)
        aL_min_adjusted = max((-self.MAX_VEL_LINEAR - self.linear_vel) / self.dt, aL_min)

        aS_max_adjusted = min((self.MAX_VEL_LATERAL - self.lateral_vel) / self.dt, aS_max)
        aS_min_adjusted = max((-self.MAX_VEL_LATERAL - self.lateral_vel) / self.dt, aS_min)

        aR_max_adjusted = min((self.MAX_VEL_ANGULAR - self.angular_vel) / self.dt, aR_max)
        aR_min_adjusted = max((-self.MAX_VEL_ANGULAR - self.angular_vel) / self.dt, aR_min)

        aL_samples = np.linspace(aL_min_adjusted, aL_max_adjusted, sample_num)
        aS_samples = np.linspace(aS_min_adjusted, aS_max_adjusted, sample_num)
        aR_samples = np.linspace(aR_min_adjusted, aR_max_adjusted, sample_num)

        predicted_path_to_draw = []

        dist_goal_costs = []
        dist_obstacle_costs = []
        path_keeping_costs = []
        time_to_goal_costs = []
        heading_stability_costs = []
        colregs_cost = []
        text = ' '
        warning_condition = True

        nautical_theta = math_to_nautical_angle(self.theta)
        global_vx = self.linear_vel * math.sin(nautical_theta) + self.lateral_vel * math.cos(nautical_theta)
        global_vy = self.linear_vel * math.cos(nautical_theta) - self.lateral_vel * math.sin(nautical_theta)
        U = np.array([global_vx, global_vy])

        obstacles_info = {}
        for obstacle in obstacles:
            obstacle_id = obstacle['id']

            pos_x = obstacle['pos_x']
            pos_y = obstacle['pos_y']
            shiptype = obstacle['shiptype']
            dx = pos_x - self.x
            dy = pos_y - self.y
            distance = math.sqrt(dx ** 2 + dy ** 2)
            nautical_theta = math_to_nautical_angle(math.atan2(U[1], U[0]))

            u2 = np.array([obstacle['vx'], obstacle['vy']])

            if distance <= 4 * 1.25:
               
                ar, fe = calculate_ar_fe(self.x, self.y, nautical_theta, pos_x, pos_y, obstacle['heading'])
                et, own_action, tar_action = determine_encounter_type(ar, fe)

                obstacles_info[obstacle_id] = {
                    'pos_x': pos_x,
                    'pos_y': pos_y,
                    'distance': distance,
                    'nautical_theta': nautical_theta,
                    'ar': ar,
                    'fe': fe,
                    'et': et,
                    'own_action': own_action,
                    'tar_action': tar_action
                }

                target_ship_instance = onto.Target_ships(obstacle_id)
                target_ship_instance.shiptype = shiptype
                onto.OS_MASS.involve_risk_of_collision.append(target_ship_instance)

                target_ship_instance.encounter_scenario.append(onto.scenario2)

                if et == "Head-on":
                    onto.OS_MASS.in_head_on.append(target_ship_instance)
                elif et == "Crossing":
                    onto.OS_MASS.in_crossing.append(target_ship_instance)
                elif et == "Overtaking":
                    onto.OS_MASS.in_overtaking.append(target_ship_instance)

                if own_action == "give-way":
                    onto.OS_MASS.hasCArole.append(onto.give_way)
                elif own_action == "stand-on":
                    onto.OS_MASS.hasCArole.append(onto.stand_on)

                target_ship_instance.OT_distance = distance


        for i, (aL, aS, aR) in enumerate(product(aL_samples, aS_samples, aR_samples)):
            # print(aL, aS, aR)
            vLpossible = self.linear_vel + aL * self.dt
            vSpossible = self.lateral_vel + aS * self.dt
            vRpossible = self.angular_vel + aR * self.dt

            predict_x, predict_y, predict_theta, path = self.predict_position(vLpossible, vSpossible,
                                                                              vRpossible, self.TAU)
            predicted_path_to_draw.append(path)

            # 1. Cost of reaching the goal
            previousTargetDistance = math.sqrt((self.x - goal[0]) ** 2 + (self.y - goal[1]) ** 2)
            newTargetDistance = math.sqrt((predict_x - goal[0]) ** 2 + (predict_y - goal[1]) ** 2)

            dist_goal_costs.append((previousTargetDistance - newTargetDistance))

            # 2. Obstacle cost
            distanceToObstacle = self.calculateClosestObstacleDistance(predict_x, predict_y, obstacles, self.TAU)

            # dcpaToObstacle = self.calculateClosestObstacleDistance(predict_x, predict_y, obstacles, self.TAU)
            # obstacle_cost = ((self.SAFE_DIST - distanceToObstacle)) if distanceToObstacle < self.SAFE_DIST else 0
            dist_obstacle_costs.append(distanceToObstacle)
            # dist_obstacle_costs = self.calculateClosestObstacleDistance(predict_x, predict_y, obstacles, self.TAU)

            # 3. Path-keeping cost
            path_vector = np.array([goal[0] - self.init_pose[0], goal[1] - self.init_pose[1]])

            position_vector = np.array([predict_x - self.init_pose[0], predict_y - self.init_pose[1]])
            cross_product = np.cross(path_vector, position_vector)
            path_deviation = np.linalg.norm(cross_product) / np.linalg.norm(path_vector)

            path_keeping_costs.append((1 / (1 + path_deviation)))

            # 4. Time to goal cost
            vector_to_goal = np.array([goal[0] - predict_x, goal[1] - predict_y])
            predict_theta1 = math_to_nautical_angle(predict_theta)
            predicted_velocity_vector = np.array(
                [vSpossible * math.cos(predict_theta1) + vLpossible * math.sin(predict_theta1),
                 vLpossible * math.cos(predict_theta1) - vSpossible * math.sin(predict_theta1)])
            cos_angle = np.dot(vector_to_goal, predicted_velocity_vector) / (
                    np.linalg.norm(vector_to_goal) * np.linalg.norm(predicted_velocity_vector) + 1e-10)
            time_to_goal = newTargetDistance / (np.linalg.norm(U) * cos_angle + 1e-10)
            time_to_goal_costs.append((1 / (1 + time_to_goal)))

            # 5. COLREGs cost
            colregs_cost_for_action = 0
            for obstacle_id, obstacle_info in obstacles_info.items():
                pos_x = obstacle_info['pos_x']
                pos_y = obstacle_info['pos_y']
                distance = obstacle_info['distance']

                nautical_theta = obstacle_info['nautical_theta']
                ar = obstacle_info['ar']
                fe = obstacle_info['fe']
                et = obstacle_info['et']
                own_action = obstacle_info['own_action']
                tar_action = obstacle_info['tar_action']

                enhanced_cost, text, warning_condition = self.enhanced_colregs_compliance(distance, et, own_action,
                                                                                          pos_x, pos_y, predict_x,
                                                                                          predict_y, vLpossible,
                                                                                          vSpossible, vRpossible)

                colregs_cost_for_action += enhanced_cost
            else:
                encounter_info = {}
                colregs_cost_for_action += 0

            # DCPA and TCPA cost
            detection_range = 10
            nautical_theta = math_to_nautical_angle(self.theta)

            # print(CA_ob_cost)
            colregs_cost.append(colregs_cost_for_action)

            heading_difference = self.calculate_heading_difference(self.init_pose, goal, predict_theta)
            heading_stability_costs.append(heading_difference ** 2)

        dist_goal_costs_norm = normalize_1(dist_goal_costs)
        dist_obstacle_costs_norm = normalize(dist_obstacle_costs)
        path_keeping_costs_norm = normalize(path_keeping_costs)
        time_to_goal_costs_norm = normalize(time_to_goal_costs)
        heading_stability_costs_norm = normalize(heading_stability_costs)
        colregs_costs_norm = normalize(colregs_cost)

        total_benefits = []

        for i in range(len(dist_goal_costs)):
            total_benefit = (self.FORWARDWEIGHT * dist_goal_costs_norm[i] +
                             self.PATHKEEPINGWEIGHT * path_keeping_costs_norm[i] -
                             self.OBSTACLEWEIGHT * dist_obstacle_costs_norm[i] +
                             self.TIMETOGOALWEIGHT * time_to_goal_costs_norm[i] -
                             self.HEADINGSTABILITYWEIGHT * heading_stability_costs_norm[i] +
                             self.COLREGSWEIGHT * colregs_costs_norm[i])
            total_benefits.append(total_benefit)

        best_action_index = total_benefits.index(max(total_benefits))
        bestBenifitGroup = total_benefits[best_action_index]
        # print(bestBenifitGroup)
        optimal_path = predicted_path_to_draw[best_action_index]  # 选择最优路径

        vLchosen = self.linear_vel + (aL_samples[best_action_index // (sample_num * sample_num)]) * self.dt
        vSchosen = self.lateral_vel + (aS_samples[(best_action_index // sample_num) % sample_num]) * self.dt
        vRchosen = self.angular_vel + (aR_samples[best_action_index % sample_num]) * self.dt

        # Update velocities
        self.linear_vel = vLchosen
        self.lateral_vel = vSchosen
        self.angular_vel = vRchosen

        # update ontology properties

        update_instance_properties(onto, onto.OS_MASS,
                                   data_properties={"hasRotationVelocity": round(self.angular_vel, 2),
                                                    "hasSurgeVelocity": round(self.linear_vel, 2),
                                                    "hasSwayVelocity": round(self.angular_vel, 2),
                                                    "hasRotationAcc": round(aR_samples[best_action_index % sample_num],
                                                                            2),
                                                    "hasSurgeAcc": round(
                                                        aL_samples[best_action_index // (sample_num * sample_num)], 2),
                                                    "hasSwayeAcc": round(
                                                        aS_samples[(best_action_index // sample_num) % sample_num], 2)})

        return predicted_path_to_draw, optimal_path, best_action_index, {
            'Dist goal': dist_goal_costs_norm,
            'Dist obstacle': dist_obstacle_costs_norm,
            'Path keeping': path_keeping_costs_norm,
            'Time to_goal': time_to_goal_costs_norm,
            'Heading stability': heading_stability_costs_norm,
            'COLREGs adherence': colregs_costs_norm
        }, vLchosen, vSchosen, vRchosen, text, warning_condition

    def move_USV(self):

        """ Move USV based on chosen velocities in dt time"""
        self.x, self.y, self.theta, tmp_path = self.predict_position(self.linear_vel, self.lateral_vel,
                                                                     self.angular_vel, self.dt)


if __name__ == '__main__':
    env = Environment(NewDWA, 20)
    running = True
    while running:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
            elif event.type == pygame.MOUSEBUTTONDOWN:
                if event.button == 3: 
                    env.paused = not env.paused

        if not env.paused and env.sim_times < 1:
            env.run()

    # while env.sim_times < 20:
    #     env.run()

    ## Please Uncomment the following lines to save simalution data

    env.avg_tg = sum(env.tg_vec) / len(env.tg_vec)
    env.avg_distance_travelled = sum(env.distance_travelled_vec) / len(env.distance_travelled_vec)

    # sort tg_vec
    tg_vec_sorted = sorted(env.tg_vec)
    tg_75th = tg_vec_sorted[int(0.75 * len(tg_vec_sorted)) - 1]
    tg_90th = tg_vec_sorted[int(0.90 * len(tg_vec_sorted)) - 1]

    # Save printed varibles into txt
    res_str1 = '[Collision Rate: {}/{}={:.2f}%] \n'.format(env.collision_times, env.sim_times,
                                                           env.collision_times / env.sim_times * 100)
    res_str2 = '[Average time to goal: {:.2f} secs] \t [tg_75th: {:.2f} secs] \t [tg_90th: {:.2f} secs] \n'.format(
        env.avg_tg, tg_75th, tg_90th)
    res_str3 = '[Average distance travelled to goal: \t {:.2f} meters] \n'.format(env.avg_distance_travelled)
    to_txt = res_str1 + res_str2 + res_str3
    print("\n" + "* " * 30 + "\n")
    print(to_txt)
    print("\n" + "* " * 30 + "\n")
    with open('static_goal_new_dwa_result{}.txt'.format(env.init_obstacle_num), 'w') as f:
        f.write(to_txt)

    # Save tg_vec into csv file
    tg_file = pd.DataFrame(data=env.tg_vec, columns=["tg"])
    tg_file.to_csv("static_goal_new_dwa_tg{}.csv".format(env.init_obstacle_num))
    vel_vec = [[env.vl_vec[idx], env.vs_vec[idx], env.vr_vec[idx]] for idx in range(len(env.vl_vec))]
    vel_file = pd.DataFrame(data=vel_vec, columns=["vl", "vs", "vr"])
    vel_file.to_csv("static_goal_new_dwa_vel{}.csv".format(env.init_obstacle_num))

    pygame.quit()

    # Plot
    plt.figure()
    plt.plot(list(range(len(env.tg_vec))), env.tg_vec, 'bo', list(range(len(env.tg_vec))), env.tg_vec, 'k')
    plt.title("time to goal")
    plt.xlabel("simulation times")
    plt.ylabel("tg(sec)")
    plt.figure()
    plt.subplot(311)
    plt.plot(list(range(len(env.vl_vec))), env.vl_vec, 'r', label="linear velocity")
    plt.xlabel("simulation steps")
    plt.ylabel("vl(m/s)")
    plt.legend(loc="upper right")
    plt.subplot(312)
    plt.plot(list(range(len(env.vl_vec))), env.vs_vec, 'b', label="lateral velocity")
    plt.xlabel("simulation steps")
    plt.ylabel("vs(m/s)")
    plt.legend(loc="upper right")

    plt.subplot(313)
    plt.plot(list(range(len(env.vr_vec))), env.vr_vec, 'g', label="angular velocity")
    plt.xlabel("simulation steps")
    plt.ylabel("vr(rad/s)")
    plt.legend(loc="upper right")
    plt.show()
