import torch
import torch.nn as nn
from torch.utils.data import DataLoader, TensorDataset
from sklearn.preprocessing import LabelEncoder, MinMaxScaler
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
import numpy as np
import pandas as pd
from torch.utils.data import Dataset, DataLoader

df = pd.read_csv(file_name)

encoder = OneHotEncoder(sparse=False)
categories = df[['Risk_Level', 'encounter_type', 'own_ship_role', 'tar_ship_role']]
categories_encoded = encoder.fit_transform(categories)
categories_encoded_df = pd.DataFrame(categories_encoded, columns=encoder.get_feature_names_out())

df_rest = df.drop(['Risk_Level', 'encounter_type', 'own_ship_role', 'tar_ship_role'], axis=1).reset_index(drop=True)
df_encoded = pd.concat([df_rest, categories_encoded_df], axis=1)



class CollisionSequenceDataset(Dataset):
  def __init__(self, df, sequence_length=30, feature_columns=None, label_column='Ownship_Behavior'):
    self.df = df
    self.sequence_length = sequence_length
    self.feature_columns = feature_columns
    self.label_column = label_column
    self.scaler = MinMaxScaler()
    self.label_encoder = LabelEncoder()
    self.prepare_data()

  def prepare_data(self):
    self.sequences = []
    self.labels = []

    for _, group in self.df.groupby('collision_key'):
      group = group.sort_values(by='time')
      if self.feature_columns is None:
        self.feature_columns = [col for col in group.columns if
                                col not in ['collision_key', 'time', self.label_column, 'Risk_Level']]
      features = self.scaler.fit_transform(group[self.feature_columns])
      labels = self.label_encoder.fit_transform(group[self.label_column])

      # Padding
      for start_idx in range(0, len(features), self.sequence_length):
        end_idx = start_idx + self.sequence_length
        seq = features[start_idx:end_idx]
        if len(seq) < self.sequence_length:
          seq = np.vstack((np.zeros((self.sequence_length - len(seq), seq.shape[1])), seq))
        self.sequences.append(seq)
        self.labels.append(labels[min(end_idx, len(features) - 1)])

    self.sequences = torch.tensor(self.sequences, dtype=torch.float)
    self.labels = torch.tensor(self.labels, dtype=torch.long)

  def __len__(self):
    return len(self.sequences)

  def __getitem__(self, idx):
    return self.sequences[idx], self.labels[idx]

sequences = np.array(sequences)
labels = np.array(labels)


def pad_sequences(sequences, maxlen, padding='post', truncating='post', value=0):

  padded_sequences = []
  for seq in sequences:
    if len(seq) < maxlen:
      # Need to pad
      pad_size = maxlen - len(seq)
      if padding == 'pre':
        pad_seq = [value] * pad_size + seq
      else:  # 'post'
        pad_seq = seq + [value] * pad_size
      padded_sequences.append(pad_seq)
    else:
      # Need to truncate
      if truncating == 'pre':
        padded_sequences.append(seq[-maxlen:])
      else:  # 'post'
        padded_sequences.append(seq[:maxlen])
  return np.array(padded_sequences)



class ShipBehaviorLSTM(nn.Module):
  def __init__(self, input_size, hidden_layer_size, output_size, num_layers):
    super(ShipBehaviorLSTM, self).__init__()
    self.hidden_layer_size = hidden_layer_size

    self.lstm = nn.LSTM(input_size, hidden_layer_size, num_layers, batch_first=True)

    self.linear = nn.Linear(hidden_layer_size, output_size)

    self.hidden_cell = (torch.zeros(num_layers, 1, self.hidden_layer_size),
                        torch.zeros(num_layers, 1, self.hidden_layer_size))

  def forward(self, input_seq):
    lstm_out, self.hidden_cell = self.lstm(input_seq.view(len(input_seq), 1, -1), self.hidden_cell)
    predictions = self.linear(lstm_out.view(len(input_seq), -1))
    return predictions[-1]

input_size = len(feature_columns) 
hidden_layer_size = 100
output_size = len(np.unique(encoded_labels)) 
num_layers = 1

model = ShipBehaviorLSTM(input_size, hidden_layer_size, output_size, num_layers)


loss_function = nn.CrossEntropyLoss()  
optimizer = torch.optim.Adam(model.parameters(), lr=0.001)



epochs = 10 
for epoch in range(epochs):
    for seq, labels in zip(X_train_tensor, y_train_tensor):
        optimizer.zero_grad() 
        model.hidden_cell = (torch.zeros(num_layers, 1, model.hidden_layer_size),
                             torch.zeros(num_layers, 1, model.hidden_layer_size)) 

        y_pred = model(seq)  

        single_loss = loss_function(y_pred, labels.long())
        single_loss.backward()
        optimizer.step() 

    if epoch%2 == 1:
        print(f'epoch: {epoch:3} loss: {single_loss.item():10.8f}')